/**
	@file nscc.d
	@brief NScripter向けのファイル名変換機能を提供する。
	@par 謝辞:
		@li for bogoyaml.d<br/>
			shinichiro.hさんのbogoYAMLを利用させていただいてます。<br/>
			URL[http://shinh.skr.jp/bogoyaml/]
			なお、警告を排除するため、少々変更をしています。<br/>
**/
module nscc.nscc;

private import std.file;
private import std.string;
private import std.path;
private import std.stdio;
private import std.conv;
private import std.windows.charset;
private import std.compiler;

private import bogoyaml;

private import util.filetime;

static this(){
	VERSION_STR ~= std.path.linesep ~ 
					"COMPILED BY: " ~ std.compiler.name ~ 
					" Ver. " ~ std.string.toString(std.compiler.version_major) ~ "." ~ std.string.toString(std.compiler.version_minor) ~
					" supported D spec " ~ std.string.toString(std.compiler.D_major) ~ "." ~ std.string.toString(std.compiler.D_minor);
}

/// バージョン情報
const char[] VERSION = "1.00";
/// バージョン文字列
char[] VERSION_STR = "NSCC VERSION: " ~ VERSION;
/// USAGE文字列
const char[] USAGE_STR = "
nscc [ファイル名[:番号]]... [-d:出力ディレクトリ名] [-ext:拡張子] [-doextract] [-doifmodified] [-env:設定ファイル名]

ファイル名   ：変換対象のファイル名です。複数の場合は半角空白区切りで指定します。
番号         ：変換先の番号です。「番号.txt」のように変換されます。
               ファイル名の直後に指定してください。
               例) nscc define.nsc:0 scenario.nsc:1
-d           ：出力先のディレクトリを指定します。デフォルトは「.」です。
-ext         ：抽出時の拡張子を指定します。デフォルトは「nsc」です。
-doextract   ：強制的に抽出するようにします。
               この指定なしでファイル名を指定すると、抽出は行われません。
-doifmodified：更新時のみ変換するようにします。
               変換先のファイルがない場合も変換します。
-setting     ：設定ファイルを指定します。デフォルトは「nsccenv.yml」です。
               変換対象ファイル指定時に設定ファイルを読み込ませたい場合にも指定します。
";
/// 最大ファイル数
const int MAX_NUMBER_OF_FILE = 100;
/// ファイル無指定
const int FILE_NUMBER_UNDEFINED = -1;
/// デフォルト設定ファイル名
const char[] SETTING_FILE_DEFAULT = "nsccenv.yml";

alias char[][char[]] Properties;
alias NSCC_ NSCCEnv;
alias std.windows.charset.toMBSz toPFCharSetz;
alias std.windows.charset.fromMBSz fromPFCharSetz;

/// USAGE例外
class UsageException : Exception
{
	this(){
		super(USAGE_STR);
	}
}

/// VERSION例外
class VersionException : Exception
{
	this(){
		super(VERSION_STR);
	}
}

/// Setting例外
class SettingException : Exception
{
	this(char[] msg){
		super("Invalid Setting File - " ~ msg);
	}
}

/// NSCC の実体
struct NSCC_
{
	int[char[]] src_files; ///< 変換元となるファイル+番号指定
	bool doextract = false; ///< 抽出実行するかどうか
	char[] src_dir = "."; ///< 抽出対象となるディレクトリ
	char[] dst_dir = "."; ///< 出力先のディレクトリ
	char[] src_ext = "nsc"; ///< 抽出対象となる拡張子
	bool doifmodified = false; ///< 更新時のみ実行
	char[] setting_file = SETTING_FILE_DEFAULT; ///< 設定ファイル名
	static NSCC_ init = {
		src_files.init, 
		doextract.init, 
		src_dir.init, 
		dst_dir.init, 
		src_ext.init,
		doifmodified.init,
		setting_file.init
		};
	void appendSrcFile(char[] src_file, int file_num=FILE_NUMBER_UNDEFINED){
		src_files[src_file] = file_num;
	}
}


/// コマンドラインを解析する。
/**
	@param args コマンドライン
	@return 解析結果
**/
Properties parseArgs(char[][] args){
	Properties properties;
	if(args.length < 2){
		return properties;
	}
	foreach(char[] arg; args[1..length]){
		char[][] factors;
		// オプション
		if(isOption(arg)){
			if(arg.length > 1 && (arg[0..1] == "-d" || arg[0..7] == "-setting") ){
				// 出力先ディレクトリ、設定ファイルの場合は切り分ける。
				factors = separateFileAndNumber(fromPFCharSetz(arg ~ "\0"));
			}
			else{
				// 2つ以上":"があっても後ろは無視
				factors = split(arg, ":");
			}
		}
		// ファイル名
		else{
			// ここで日本語の場合は変換しておく
			factors = separateFileAndNumber(fromPFCharSetz(arg ~ "\0"));
		}
		
		// プロパティに設定
		if(factors.length == 1){
			properties[factors[0]] = "";
		}
		else{
			properties[factors[0]] = factors[1];
		}
	}
	return properties;
}


/// 引数がオプションかどうか確認する。
/**
	@param arg コマンドライン引数
	@retval !=0 引数はオプション
	@retval ==0 引数はオプションではない
**/
int isOption(char[] arg){
	return ((arg.length > 0) && (arg[0] == '-'));
}


/// ファイル名:番号を切り分ける。
/**
	@param arg ファイル:番号の文字列
	@return ファイル名、番号の順に格納した配列
**/
char[][] separateFileAndNumber(char[] arg){
	char[][] factors = split(arg, ":");
	char[] file;
	char[] num;
	// ドライブ区切り文字との衝突を回避
	if(arg.length > 3 && arg[1] == ':' && arg[2] == '\\'){
		switch(factors.length)
		{
		// <Drive>:\<Path>
		case 2:
			file = arg;
			break;
		// <Drive>:\<Path>:<Number>
		case 3:
			file = factors[0] ~ ":" ~ factors[1];
			num = factors[2];
			break;
		default:
			break;
		}
	}
	else{
		switch(factors.length)
		{
		// <Path>
		case 1:
			file = arg;
			break;
		// <Path>:<Number>
		case 2:
			file = factors[0];
			num = factors[1];
			break;
		default:
			break;
		}
	}
	// 指定がおかしいとみなし、USAGE例外
	if(file == null){
		throw new UsageException;
	}
	char[][] ret;
	ret.length = 2;
	ret[0] = file;
	ret[1] = num;
	return ret;
}


/// プロパティから環境を生成する。
/**
	@param properties プロパティ
	@return NSCCの環境情報
**/
NSCCEnv makeEnvironment(Properties properties){
	NSCCEnv env;
	foreach(char[] key, char[] value; properties){
		// オプションごとに環境設定
		switch(key){
		// ディレクトリ
		case "-d":
			env.dst_dir = value;
			break;
		// 拡張子
		case "-ext":
			// 先頭'.'を許容する。
			env.src_ext = (value[0] == '.') ? value[1..length] : value;
			break;
		// 更新時に実行
		case "-doifmodified":
			// キーのみならtrue
			env.doifmodified = (value.length > 0) ? toBool(value) : true;
			break;
		// 抽出モード
		case "-doextract":
			// キーのみならtrue
			env.doextract = (value.length > 0) ? toBool(value) : true;
			break;
		// 設定ファイル
		case "-setting":
			// とりあえず格納する。
			env.setting_file = value;
			break;
		default:
			if(isOption(key)){
				// 不明なオプションはUSAGE例外かVERSION例外
				if(key == "-version"){
					throw new VersionException();
				}
				throw new UsageException();
			}
			// ファイル名:番号で環境に設定
			if(value.length > 0){
				env.appendSrcFile(key, toInt(value));
			}
			else{
				env.appendSrcFile(key);
			}
			break;
		}
	}
	return env;
}


/// 環境に応じて実行する。
/**
	@param NSCCの環境情報
**/
void execute(NSCCEnv env){
	careDir(env.dst_dir);
	if(env.doextract){
		// ファイルを抽出して追加。
		char[][] files = extractFiles(env.src_dir, env.src_ext, env.src_files.keys);
		foreach(char[] file; files){
			env.appendSrcFile(file);
		}
	}
	execute(env.src_files, env.dst_dir, env.doifmodified);
}


/// ディレクトリが無ければ作成する。
/**
	@param dir ディレクトリ名
**/
void careDir(char[] dir){
	if(dir != std.file.curdir){
		if(!exists(dir)){
			mkdir(dir);
		}
		else if(!isdir(dir)){
			throw new Exception("Can Not Care Directory");
		}
	}
}


/// excludesを除外したファイルを抽出する。
/**
	@param dir 抽出対象ディレクトリ
	@param ext 抽出対象拡張子
	@param excludes 除外するファイル名
	@return ファイル名のリスト
**/
char[][] extractFiles(char[] dir, char[] ext, char[][] excludes)
{
	char[][] files = listdir(dir, "*" ~ ext);
	char[][] ret;
	foreach(char[] file; files){
		if(inArray!(char[])(file, excludes)){
			continue;
		}
		ret.length = ret.length + 1;
		ret[length-1] = file;
	}
	return ret;
}


/// ファイルの変換を実行する。
/**
	@param src_files ソースファイル名と番号のリスト
	@param dst_dir 変換先のディレクトリ
	@param doifmodified falseなら常に、trueなら更新時のみ実行する。
**/
void execute(int[char[]] src_files, char[] dst_dir, bool doifmodified){
	assert(isdir(dst_dir));
	assert(src_files.length > 0);
	
	// 使用する変数を初期化する。
	bool[MAX_NUMBER_OF_FILE] used_num;
	int file_num = 0;
	
	// 番号指定のファイルを予約しておく。
	foreach(int reserved; src_files.values){
		if(reserved != FILE_NUMBER_UNDEFINED){
			used_num[reserved] = true;
		}
	}
	// ファイルごとに処理
	foreach(char[] src_file, int file_num; src_files){
		// ここで存在しないファイルは処理のしようが無い。
		assert(isfile(src_file));
		
		// 無指定なら未使用番号を取得
		if(file_num == FILE_NUMBER_UNDEFINED){
			file_num = getUnusedNumber(used_num);
			// 未使用番号も無ければエラー
			if(file_num == FILE_NUMBER_UNDEFINED){
				throw new Exception("No Number To Use");
			}
		}
		// 番号が最大値を超過したらエラー
		if(file_num >= MAX_NUMBER_OF_FILE){
			throw new Exception("Over Max File Number");
		}
		// 使用済み
		used_num[file_num] = true;
		// .txt に出力
		char[] dst_file = dst_dir ~ std.path.sep ~ zfill(std.string.toString(file_num), 2) ~ ".txt";
		if(doifmodified && exists(dst_file)){
			// 変換対象と変換先が保持しているタイムスタンプを比較し、変換対象のタイムスタンプが等しいか古ければ変換しない。
			if(compareTimeStamp(src_file, dst_file) <= 0){
				// ログ
				writefln("%s is not newer than %s", getBaseName(src_file), getBaseName(dst_file));
				continue;
			}
		}
		// 変換対象のファイルを、変換先へとコピーする。
		copy(src_file, dst_file);
		syncTimeStamp(src_file, dst_file);
		// ログ
		printf("%s => %s\n", toPFCharSetz(getBaseName(src_file)), toPFCharSetz(getBaseName(dst_file)));
	}
}


/// 変換先番号を取得する。
/**
	@param used_num trueが使用中を表すリスト
	@return 変換先番号
**/
int getUnusedNumber(bool[] used_num){
	int ret = FILE_NUMBER_UNDEFINED;
	foreach(int num, bool used; used_num){
		if(!used){
			ret = num;
			break;
		}
	}
	return ret;
}


/// タイムスタンプを比較する。
/**
	@param used_num trueが使用中を表すリスト
	@return 変換先番号
**/
int compareTimeStamp(char[] src_file, char[] dst_file){
	assert(exists(src_file));
	assert(exists(dst_file));
	FileTime src_time = getfiletime(src_file);
	FileTime dst_time = getfiletime(dst_file);
	
	return (cast(uint)src_time.lastWrite - cast(uint)dst_time.lastWrite);
}


/// タイムスタンプの同期を取る。
/**
	@param src_file 同期元ファイル
	@param dst_file 同期先ファイル。アクセス日時のみ同期にならない。
**/
void syncTimeStamp(char[] src_file, char[] dst_file){
	assert(exists(src_file));
	assert(exists(dst_file));
	FileTime src_time = getfiletime(src_file);
	setfiletime(dst_file, src_time);
}


/// 設定ファイルを読み込む。
/**
	@par 想定
	@code
	---
	nscc:
		includes:
			- ファイル[: 番号]
			[- 次のファイル[: 番号]]
		[d: ディレクトリ名]
		[ext: 拡張子]
		# 以下は必ず値を指定すること。
		[doextract: true|false]
		[doifmodified: true|false]
	@endcode
**/
Properties loadSetting(char[] name){
	if(!exists(name)){
		return null;
	}
	Properties props;
	try{
		YamlHolder yh_root = yamlLoad(name);
		if(!yh_root.get("nscc").isAssoc()){
			return null;
		}
		printf("loaded\n");
		YamlHolder[char[]] yh_nscc = yh_root.get("nscc").assoc;
		printf("get nscc\n");
		foreach(char[] key, YamlHolder yh_prop; yh_nscc){
			printf("get prop %.*s\n", key);
			switch(key){
			case "includes":
				// 記法を統一するため、配列のみを考慮
				char[] src_file;
				char[] dst_num;
				char[][] factors;
				foreach(YamlHolder yh_file; yh_prop.array){
					if(yh_file.isScalar()){
						factors = separateFileAndNumber(fromPFCharSetz(yh_file.str ~ "\0"));
						src_file = factors[0];
						dst_num = factors[1];
					}
					// ここで配列はありえない
					else{
						throw new SettingException("cannot use array for indicating a factor of includes");
					}
					printf("includes %.*s:%.*s\n", src_file, dst_num);
					props[src_file] = dst_num;
				}
				break;
			case "doifmodified":
			case "doextract":
			case "d":
			case "ext":
				// 複数の値をとらないのでスカラーしかない
				if(!yh_prop.isScalar()){
					throw new SettingException("cannot use array or assoc for indicating a factor of " ~ key);
				}
				props["-" ~ key] = yh_prop.str;
				break;
			default:
				// 不正なオプションは無視
				break;
			}
		}
	}
	catch(YamlError e){
		throw new SettingException("bad YAML format" ~ std.path.linesep ~ e.toString());
	}
	return props;
}

unittest
{
	assert(fromMBSz("\x93\xfa\x96\x7b\x8c\xea\0") == "日本語");
	debug printf("test isOption...");
	assert(isOption("-option"));
	assert(!isOption("option"));
	debug printf("passed\n");
	
	debug printf("test separateFileAndNumber...");
	char[][] factor;
	factor = separateFileAndNumber("file.txt");
	assert(factor.length == 2);
	assert(factor[0] == "file.txt");
	factor = separateFileAndNumber("file.txt:0");
	assert(factor.length == 2);
	assert(factor[0] == "file.txt");
	assert(factor[1] == "0");
	factor = separateFileAndNumber(r"c:\file.txt");
	assert(factor.length == 2);
	assert(factor[0] == r"c:\file.txt");
	factor = separateFileAndNumber(r"c:\file.txt:0");
	assert(factor.length == 2);
	assert(factor[0] == r"c:\file.txt");
	assert(factor[1] == "0");
	debug printf("passed\n");
	
	// 正常パラメータ
	static char[][][] argss = [
		[ "nscc.exe", "-d:debug", "-ext:nss" ],
		[ "nscc.exe", r"c:\file dir\file.nsc", "file.txt", "-d:specified"],
		[ "nscc.exe", r"c:\file dir\file.nsc:20", "file.txt:15" ],
		[ "nscc.exe", "define.nsc:99", "entry.txt:0", "-d:debug", "-ext:nss", "-doextract", "-doifmodified" ]
	];
	debug printf("test parseArgs...");
	Properties props;
	props = parseArgs(argss[0]);
	assert(props["-d"] == "debug");
	assert(props["-ext"] == "nss");
	props = parseArgs(argss[1]);
	assert(props[r"c:\file dir\file.nsc"] == "");
	assert(props["file.txt"] == "");
	assert(props["-d"] == "specified");
	props = parseArgs(argss[2]);
	assert(props[r"c:\file dir\file.nsc"] == "20");
	assert(props["file.txt"] == "15");
	props = parseArgs(argss[3]);
	assert(props["define.nsc"] == "99");
	assert(props["entry.txt"] == "0");
	assert(props["-d"] == "debug");
	assert(props["-ext"] == "nss");
	assert(props["-doextract"] == "");
	assert(props["-doifmodified"] == "");
	debug printf("passed\n");

	debug printf("test makeEnvironment...");
	NSCCEnv env = makeEnvironment(parseArgs(argss[0]));
	assert(env.src_dir == ".");
	assert(env.dst_dir == "debug");
	assert(env.src_ext == "nss");
	env = makeEnvironment(parseArgs(argss[1]));
	assert(env.dst_dir == "specified");
	assert(r"c:\file dir\file.nsc" in env.src_files);
	assert("file.txt" in env.src_files);
	env = makeEnvironment(parseArgs(argss[2]));
	assert(env.src_files[r"c:\file dir\file.nsc"] == 20);
	assert(env.src_files["file.txt"] == 15);
	env = makeEnvironment(parseArgs(argss[3]));
	assert(env.src_files["define.nsc"] == 99);
	assert(env.src_files["entry.txt"] == 0);
	assert(env.src_dir == ".");
	assert(env.dst_dir == "debug");
	assert(env.src_ext == "nss");
	assert(env.doextract);
	assert(env.doifmodified);
	debug printf("passed\n");
	
	// loadSetting
	debug printf("test loadSetting...");
	props = loadSetting("nsccenv.yml"); // test file
	assert(props["loadtest.nss"] == "0");
	assert(props["loadtest1.nsc"] == "1");
	assert(props["-d"] == "loadtest");
	assert(props["-ext"] == "nsy");
	assert(props["-doifmodified"] == "false");
	assert(props["-doextract"] == "true");
	debug printf("passed\n");
}

/// メインクラス
class NSCC
{
	this(){
	}
	
	/// 保持しているデータを初期化する。
	void init(){
		self = self.init;
	}
	
	/// コマンドラインから生成する。
	static NSCC createByCommandLine(char[][] args){
		Properties properties = parseArgs(args);
		// 設定ファイルの指定があれば、それも読み込む。
		if("-setting" in properties){
			char[] setting_file = (properties["-setting"].length > 0) ? properties["-setting"] : SETTING_FILE_DEFAULT;
			Properties settings = loadSetting(setting_file);
			// 設定ファイルより、コマンドライン指定を優先する。
			foreach(char[] key, char[] value; settings){
				if(key in properties){
					continue;
				}
				properties[key] = value;
			}
		}
		return createByProperties(properties);
	}
	
	/// 設定ファイルから生成する。
	static NSCC createBySettingFile(char[] name){
		Properties properties = loadSetting(name);
		return createByProperties(properties);
	}
	
	/// プロパティから生成する。
	static NSCC createByProperties(Properties properties){
		NSCC nscc = new NSCC();
		nscc.self = makeEnvironment(properties);
		return nscc;
	}

	/// ファイルを変換する。
	void execute(){
		nscc.nscc.execute(self);
	}

private:
	/// NSCCの実体
	NSCC_ self;
}

unittest
{
	debug(nscc) printf("nscc.nscc");

	// デフォルトパラメータ
	debug(nscc) printf("test: default parameter...");
	NSCC nscc = new NSCC();
	with(nscc.self){
		assert(dst_dir == ".");
		assert(src_dir == ".");
		assert(src_ext == "nsc");
	}
	debug(nscc) printf("passed\n");
	
	// 正常パラメータ
	static char[][][] argss = [
		[ "nscc.exe", "-d:debug", "-ext:nss" ],
		[ "nscc.exe", r"c:\file dir\file.nsc", "file.txt", "-d:specified"],
		[ "nscc.exe", r"c:\file dir\file.nsc:20", "file.txt:15" ],
	];
	
	// オプション指定
	debug(nscc) printf("test: option parameter...");
	nscc = NSCC.createByCommandLine(argss[0]);
	with(nscc.self){
		assert(dst_dir == "debug");
		assert(src_ext == "nss");
	}
	debug(nscc) printf("passed\n");
	
	// ファイル指定
	debug(nscc) printf("test: file specified...");
	nscc = NSCC.createByCommandLine(argss[1]);
	with(nscc.self){
		assert(dst_dir == "specified");
		assert(src_ext == "nsc");
		assert(r"c:\file dir\file.nsc" in src_files);
		assert("file.txt" in src_files);
	}
	debug(nscc) printf("passed\n");
	
	// ファイル:番号指定
	debug(nscc) printf("test: file and number specified...");
	nscc = NSCC.createByCommandLine(argss[2]);
	with(nscc.self){
		assert(src_files.length == 2);
		assert(src_files[r"c:\file dir\file.nsc"] == 20);
		assert(src_files["file.txt"] == 15);
	}
	debug(nscc) printf("passed\n");
	
	return 0;

}

/// 文字列を真偽値に変換する。
bool toBool(char[] s)
{
	bool ret;
	switch(tolower(s)){
	case "enable":
	case "true":
	case "on":
	case "yes":
	case "ok":
	case "1":
		ret = true;
		break;
	default:
		ret = false;
		break;
	}
	return ret;
}

unittest
{
	debug(util) printf("test: toBool...");
	assert(toBool("true"));
	assert(toBool("on"));
	assert(toBool("yes"));
	assert(toBool("1"));
	assert(!toBool("false"));
	assert(!toBool("off"));
	assert(!toBool("no"));
	assert(!toBool("0"));
	debug(util) printf("passed\n");
}

template inArray(T)
{
	bool inArray(T one, T[] array){
		bool ret;
		foreach(T factor; array){
			if(one == factor){
				ret = true;
				break;
			}
		}
		return ret;
	}
}

unittest
{
	debug(util) printf("test: inArray...");
	static char[][] str_array = ["hoge", "fuga", "foo", "bar"];
	assert(inArray!(char[])("hoge", str_array));
	assert(inArray!(char[])("fuga", str_array));
	assert(inArray!(char[])("foo", str_array));
	assert(inArray!(char[])("bar", str_array));
	assert(!inArray!(char[])("nai", str_array));
	debug(util) printf("passed\n");
}
