module util.filetime;

import std.date;

private import std.file;
private import std.c.windows.windows;
private import std.windows.syserror;
private import std.c.stdio;
private import std.c.time;

bool useWAPI = true;

static this()
{
    // do not use the W APIs in case of Win 95, 98, Me.
    useWAPI = (GetVersion() < 0x80000000);
}

struct TimeStamp
{
	void opCall(Date aDate){
		opCall(aDate.year, aDate.month, aDate.day, aDate.hour, aDate.minute, aDate.second, aDate.ms, aDate.weekday);
	}
	void opCall(int aYear, int aMonth, int aDay, int aHour=0, int aMinute=0, int aSecond=0, int aMilliSeconds=0, int aWeekDay=0){
		tm temp;
		temp.tm_sec   = date.second  = aSecond;
		temp.tm_min   = date.minute  = aMinute;
		temp.tm_hour  = date.hour    = aHour;
		temp.tm_mday  = date.day     = aDay;
		temp.tm_mon   = date.month   = aMonth;
		temp.tm_year  = date.year    = aYear;
		temp.tm_wday  = date.weekday = aWeekDay;
		temp.tm_yday  = 0;
		temp.tm_isdst = 0;
		value = mktime(&temp) + aMilliSeconds;
		date.ms = aMilliSeconds;
	}
	void opCall(TimeStamp aTimeStamp){
		date = aTimeStamp.date;
		value = aTimeStamp.value;
	}
	uint opCast(){
		return value;
	}
	Date date;
	uint value;
}

unittest
{
	TimeStamp ts1, ts2, ts3;
	Date date;
	date.parse("2006/06/21");
	ts1(2006, 6, 21);
	ts2(ts1);
	ts3(date);
	assert(ts1 == ts2);
	assert(ts1 == ts3);
}


struct FileTime
{
	TimeStamp creation;
	TimeStamp lastAccess;
	TimeStamp lastWrite;
}

//version (Win32)

export
{
	extern(Windows) BOOL GetFileTime(HANDLE hFile, LPFILETIME lpCreationTime, LPFILETIME lpLastAccessTime, LPFILETIME lpLastWriteTime);
	extern(Windows) BOOL SetFileTime(HANDLE hFile, FILETIME* lpCreationTime, FILETIME* lpLastAccessTime, FILETIME* lpLastWriteTime);
}

FileTime getfiletime(char[] name)
{
	HANDLE hFile;
	FILETIME filetime_creation;
	FILETIME filetime_last_access;
	FILETIME filetime_last_write;
	SYSTEMTIME systemtime_creation;
	SYSTEMTIME systemtime_last_access;
	SYSTEMTIME systemtime_last_write;

	hFile = GetFileHandle(name);
	if(hFile == INVALID_HANDLE_VALUE){
		throw new FileException(name, GetLastError());
	}
	if(!GetFileTime(hFile , &filetime_creation, &filetime_last_access, &filetime_last_write)){
		CloseHandle(hFile);
		throw new FileException(name, GetLastError());
	}

	CloseHandle(hFile);
	if(!FileTimeToLocalSystemTime(&filetime_creation , &systemtime_creation)){
		throw new FileException(name, GetLastError());
	}
	if(!FileTimeToLocalSystemTime(&filetime_last_access , &systemtime_last_access)){
		throw new FileException(name, GetLastError());
	}
	if(!FileTimeToLocalSystemTime(&filetime_last_write , &systemtime_last_write)){
		throw new FileException(name, GetLastError());
	}

	FileTime ft;
	ft.creation(SystemTimeToDate(&systemtime_creation));
	ft.lastAccess(SystemTimeToDate(&systemtime_last_access));
	ft.lastWrite(SystemTimeToDate(&systemtime_last_write));
	return ft;
}


void setfiletime(char[] name, FileTime ft)
{
	HANDLE hFile;
	FILETIME filetime_creation;
	FILETIME filetime_last_access;
	FILETIME filetime_last_write;
	SYSTEMTIME systemtime_creation;
	SYSTEMTIME systemtime_last_access;
	SYSTEMTIME systemtime_last_write;

	hFile = GetFileHandle(name, true);
	if(hFile == INVALID_HANDLE_VALUE){
		throw new FileException(name, GetLastError());
	}
	
	DateToSystemTime(ft.creation.date, &systemtime_creation);
	DateToSystemTime(ft.lastAccess.date, &systemtime_last_access);
	DateToSystemTime(ft.lastWrite.date, &systemtime_last_write);
	
	if(!SystemTimeToUTCFileTime(&systemtime_creation, &filetime_creation)){
		CloseHandle(hFile);
		throw new FileException(name, GetLastError());
	}
	if(!SystemTimeToUTCFileTime(&systemtime_last_access, &filetime_last_access)){
		CloseHandle(hFile);
		throw new FileException(name, GetLastError());
	}
	if(!SystemTimeToUTCFileTime(&systemtime_last_write, &filetime_last_write)){
		CloseHandle(hFile);
		throw new FileException(name, GetLastError());
	}

	if(!SetFileTime(hFile, &filetime_creation, &filetime_last_access, &filetime_last_write)){
		CloseHandle(hFile);
		throw new FileException(name, GetLastError());
	}
	CloseHandle(hFile);
}

// wrapper

private HANDLE GetFileHandle(char[] name, bool mode_write=false)
{
	HANDLE hFile;
	DWORD dwDesiredAccess;
	DWORD dwShareMode;
	
	dwDesiredAccess = mode_write ? GENERIC_WRITE : GENERIC_READ;
	dwShareMode = 0; // FILE_SHARE_READ?
    if(useWAPI)
    {
		wchar* namez = std.utf.toUTF16z(name);
		hFile = CreateFileW(namez, dwDesiredAccess, dwShareMode, null, OPEN_EXISTING,
		    FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, cast(HANDLE)null);
    }
    else
    {
		char* namez = toMBSz(name);
		hFile = CreateFileA(namez, dwDesiredAccess, dwShareMode, null, OPEN_EXISTING,
		    FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, cast(HANDLE)null);
    }
	return hFile;
}


private bool FileTimeToLocalSystemTime(FILETIME* ft, SYSTEMTIME* st)
{
	FILETIME ft_local;
	if(!FileTimeToLocalFileTime(ft , &ft_local)){
		return false;
	}
	if(!FileTimeToSystemTime(&ft_local , st)){
		return false;
	}
	return true;
}

private bool SystemTimeToUTCFileTime(SYSTEMTIME* st, FILETIME* ft)
{
	FILETIME ft_local;
	if(!SystemTimeToFileTime(st, &ft_local)){
		return false;
	}
	if(!LocalFileTimeToFileTime(&ft_local, ft)){
		return false;
	}
	return true;
}


private Date SystemTimeToDate(SYSTEMTIME* st)
{
	Date date;
	date.year    = st.wYear;
	date.month   = st.wMonth;
	date.day     = st.wDay;
	date.hour    = st.wHour;
	date.minute  = st.wMinute;
	date.second  = st.wSecond;
	date.ms      = st.wMilliseconds;
	date.weekday = st.wDayOfWeek;
	return date;
}


private void DateToSystemTime(Date date, SYSTEMTIME* st)
{
	st.wYear         = cast(ushort)date.year;
	st.wMonth        = cast(ushort)date.month;
	st.wDay          = cast(ushort)date.day;
	st.wHour         = cast(ushort)date.hour;
	st.wMinute       = cast(ushort)date.minute;
	st.wSecond       = cast(ushort)date.second;
	st.wMilliseconds = cast(ushort)date.ms;
	st.wDayOfWeek    = cast(ushort)date.weekday;
}


//} // Win32
